import chess
import chess.pgn
import collections
import torch
import math

def pgn_from_chessboard(board:chess.Board, white_name:str="White", black_name:str="Black") -> str:
    """
    Generate a PGN from a chess.Board.
    """
    game = chess.pgn.Game()

    # Undo all moves.
    switchyard = collections.deque()
    while board.move_stack:
        switchyard.append(board.pop())

    game.setup(board)
    node = game

    # Replay all moves.
    while switchyard:
        move = switchyard.pop()
        node = node.add_variation(move)
        board.push(move)

    # Define headers of PGN game
    game.headers["White"] = white_name
    game.headers["Black"] = black_name
    game.headers["Result"] = board.result()
    
    return str(game)

def get_material_balance(board:chess.Board, queen:int=9, rook:int=5, bishop:int=3, knight:int=3, pawn:int=1) -> int:
    white = board.occupied_co[chess.WHITE]
    black = board.occupied_co[chess.BLACK]
    return (
        pawn * (chess.popcount(white & board.pawns) - chess.popcount(black & board.pawns)) +
        knight * (chess.popcount(white & board.knights) - chess.popcount(black & board.knights)) +
        bishop * (chess.popcount(white & board.bishops) - chess.popcount(black & board.bishops)) +
        rook * (chess.popcount(white & board.rooks) - chess.popcount(black & board.rooks)) +
        queen * (chess.popcount(white & board.queens) - chess.popcount(black & board.queens))
    )

def print_pretty_tensor_Nx8x8(tensor:torch.Tensor, to_int:bool=True) -> None:
    for idx, plane in enumerate(tensor):
        print(f"############### PLANE [{idx}]")
        print_pretty_tensor_8x8(plane, to_int=to_int)
        print()

def print_pretty_tensor_8x8(tensor:torch.Tensor, to_int:bool=True) -> None:
    """
    Print a 8 by 8 tensor as a 8 by 8 grid of numbers.
    Formatted from bottom to top to resemble a chessboard.
    """
    for row in range(7, -1, -1):
        for col in range(8):
            if to_int:
                print(int(tensor[row, col].item()), end=' ')
            else:
                print(tensor[row, col].item(), end=' ')
        print()

def print_pretty_bitboard(bitboard:chess.Bitboard) -> None:
    """
    Print a bitboard as a 8 by 8 grid of 1s and 0s.
    """
    # Cast to a chess.SquareSet
    square_set = chess.SquareSet(bitboard)
    print(square_set)

def _map_cp_to_win_probability(cp:int) -> float:
    """
    Convert centipawns to probability of winning.
    """
    return 1 / (1 + 10 ** (-cp / 400))

def _map_win_probability_to_cp(win_probability:float) -> int:
    """
    Convert probability of winning to centipawns.
    """
    return -400 * math.log10(1 / win_probability - 1)

def map_mate_or_cp_to_win_probability(mate:int, cp:int) -> float:
    """
    Win probability is a value between 0 and 1 where 0 means win for black and 1 means win for white.
    If there is mate the win probability is 0 or 1 and cp is ignored.
    Incompatible mate and cp values will raise a ValueError.
    """
    if mate is not None and mate != 0:
        if mate > 0:
            return 1.0
        if mate < 0:
            return 0.0
    else:
        if cp is not None:
            return _map_cp_to_win_probability(cp)
        else:
            raise ValueError("Incompatible mate and cp values")

def chessboard_to_tensorboard_12x8x8(chessboard:chess.Board) -> torch.Tensor:
    """
    Convert a chess board to a tensor board used as input layer.
    Returns a torch tensor of shape (12, 8, 8).
    The tensor is formed by 12 planes 8 by 8 each, each layer carries specific information.
    
    For pieces of each player:
    - Plane 0 for presence of player king (1 if present, 0 if not)
    - Plane 1 for presence of player queen (1 if present, 0 if not)
    - Plane 2 for presence of player rooks (1 if present, 0 if not)
    - Plane 3 for presence of player bishops (1 if present, 0 if not)
    - Plane 4 for presence of player knights (1 if present, 0 if not)
    - Plane 5 for presence of player pawns (1 if present, 0 if not)

    - Plane 6 for presence of opponent king (1 if present, 0 if not)
    - Plane 7 for presence of opponent queen (1 if present, 0 if not)
    - Plane 8 for presence of opponent rooks (1 if present, 0 if not)
    - Plane 9 for presence of opponent bishops (1 if present, 0 if not)
    - Plane 10 for presence of opponent knights (1 if present, 0 if not)
    - Plane 11 for presence of opponent pawns (1 if present, 0 if not)
    """
    # Initialize tensor
    tensorboard = torch.zeros(12, 8, 8)

    # Get the piece map dictionary
    piece_map: dict[chess.Square, chess.Piece] = chessboard.piece_map()

    # Define the piece type and color mappings
    piece_to_plane = {
        (chess.KING, chess.WHITE): 0,
        (chess.QUEEN, chess.WHITE): 1,
        (chess.ROOK, chess.WHITE): 2,
        (chess.BISHOP, chess.WHITE): 3,
        (chess.KNIGHT, chess.WHITE): 4,
        (chess.PAWN, chess.WHITE): 5,
        (chess.KING, chess.BLACK): 6,
        (chess.QUEEN, chess.BLACK): 7,
        (chess.ROOK, chess.BLACK): 8,
        (chess.BISHOP, chess.BLACK): 9,
        (chess.KNIGHT, chess.BLACK): 10,
        (chess.PAWN, chess.BLACK): 11
    }

    # Iterate over each square on the board
    for square in chess.SQUARES:
        piece = piece_map.get(square)
        if piece:
            plane = piece_to_plane.get((piece.piece_type, piece.color))
            if plane is not None:
                row, col = divmod(square, 8)
                tensorboard[plane, row, col] = 1

    return tensorboard

def chessboard_to_tensorboard_17x8x8(chessboard:chess.Board) -> torch.Tensor:
    """
    Convert a chess board to a tensor board used as input layer.
    Returns a torch tensor of shape (17, 8, 8).
    The tensor is formed by 17 planes 8 by 8 each, each layer carries specific information.
    
    For pieces of each player:
    - Plane 0 for presence of player king (1 if present, 0 if not)
    - Plane 1 for presence of player queen (1 if present, 0 if not)
    - Plane 2 for presence of player rooks (1 if present, 0 if not)
    - Plane 3 for presence of player bishops (1 if present, 0 if not)
    - Plane 4 for presence of player knights (1 if present, 0 if not)
    - Plane 5 for presence of player pawns (1 if present, 0 if not)

    - Plane 6 for presence of opponent king (1 if present, 0 if not)
    - Plane 7 for presence of opponent queen (1 if present, 0 if not)
    - Plane 8 for presence of opponent rooks (1 if present, 0 if not)
    - Plane 9 for presence of opponent bishops (1 if present, 0 if not)
    - Plane 10 for presence of opponent knights (1 if present, 0 if not)
    - Plane 11 for presence of opponent pawns (1 if present, 0 if not)

    - Plane 12 for player castling rights king side (1 if present, 0 if not)
    - Plane 13 for player castling rights queen side (1 if present, 0 if not)
    - Plane 14 for opponent castling rights king side (1 if present, 0 if not)
    - Plane 15 for opponent castling rights queen side (1 if present, 0 if not)

    - Plane 16 for en passant square (1 if present, 0 if not) (note: there can only be one en passant square at most)
    """
    # Initialize tensor
    tensorboard = torch.zeros(17, 8, 8)

    ## P0-P11: PIECES POSITIONS

    # Get the piece map dictionary
    piece_map: dict[chess.Square, chess.Piece] = chessboard.piece_map()

    # Define the piece type and color mappings
    map_piece_to_plane = {
        (chess.KING, chess.WHITE): 0,
        (chess.QUEEN, chess.WHITE): 1,
        (chess.ROOK, chess.WHITE): 2,
        (chess.BISHOP, chess.WHITE): 3,
        (chess.KNIGHT, chess.WHITE): 4,
        (chess.PAWN, chess.WHITE): 5,
        (chess.KING, chess.BLACK): 6,
        (chess.QUEEN, chess.BLACK): 7,
        (chess.ROOK, chess.BLACK): 8,
        (chess.BISHOP, chess.BLACK): 9,
        (chess.KNIGHT, chess.BLACK): 10,
        (chess.PAWN, chess.BLACK): 11
    }

    # Iterate over each square on the board
    for square in chess.SQUARES:
        piece = piece_map.get(square)
        if piece:
            plane = map_piece_to_plane.get((piece.piece_type, piece.color))
            if plane is not None:
                row, col = divmod(square, 8)
                tensorboard[plane, row, col] = 1
    

    ## P12-P15: CASTLING RIGHTS

    # P12: player castling rights king side
    if chessboard.has_kingside_castling_rights(chess.WHITE):
        tensorboard[12, :, :] = 1
    
    # P13: player castling rights queen side
    if chessboard.has_queenside_castling_rights(chess.WHITE):
        tensorboard[13, :, :] = 1
    
    # P14: opponent castling rights king side
    if chessboard.has_kingside_castling_rights(chess.BLACK):
        tensorboard[14, :, :] = 1
    
    # P15: opponent castling rights queen side
    if chessboard.has_queenside_castling_rights(chess.BLACK):
        tensorboard[15, :, :] = 1


    ## P16: EN PASSANT SQUARE
    if chessboard.ep_square is not None:
        row, col = divmod(chessboard.ep_square, 8)
        tensorboard[16, row, col] = 1

    return tensorboard

def chessboard_to_tensorboard_29x8x8(chessboard:chess.Board, mirror_if_black:bool) -> torch.Tensor:
    """
    Convert a chess board to a tensor board used as input layer.
    Returns a torch tensor of shape (29, 8, 8).
    The tensor is formed by 29 planes 8 by 8 each, each layer carries specific information.
    
    For pieces of each player:
    - Plane 0 for presence of player king (1 if present, 0 if not)
    - Plane 1 for presence of player queen (1 if present, 0 if not)
    - Plane 2 for presence of player rooks (1 if present, 0 if not)
    - Plane 3 for presence of player bishops (1 if present, 0 if not)
    - Plane 4 for presence of player knights (1 if present, 0 if not)
    - Plane 5 for presence of player pawns (1 if present, 0 if not)

    - Plane 6 for presence of opponent king (1 if present, 0 if not)
    - Plane 7 for presence of opponent queen (1 if present, 0 if not)
    - Plane 8 for presence of opponent rooks (1 if present, 0 if not)
    - Plane 9 for presence of opponent bishops (1 if present, 0 if not)
    - Plane 10 for presence of opponent knights (1 if present, 0 if not)
    - Plane 11 for presence of opponent pawns (1 if present, 0 if not)

    - Plane 12 for player castling rights king side (1 if present, 0 if not)
    - Plane 13 for player castling rights queen side (1 if present, 0 if not)
    - Plane 14 for opponent castling rights king side (1 if present, 0 if not)
    - Plane 15 for opponent castling rights queen side (1 if present, 0 if not)

    - Plane 16 for en passant square (1 if present, 0 if not) (note: there can only be one en passant square at most)

    - Plane 17 for squares attacked by player king (1 if attacked, 0 if not)
    - Plane 18 for squares attacked by player queen (1 if attacked, 0 if not)
    - Plane 19 for squares attacked by player rooks (1 if attacked, 0 if not)
    - Plane 20 for squares attacked by player bishops (1 if attacked, 0 if not)
    - Plane 21 for squares attacked by player knights (1 if attacked, 0 if not)
    - Plane 22 for squares attacked by player pawns (1 if attacked, 0 if not)

    - Plane 23 for squares attacked by opponent king (1 if attacked, 0 if not)
    - Plane 24 for squares attacked by opponent queen (1 if attacked, 0 if not)
    - Plane 25 for squares attacked by opponent rooks (1 if attacked, 0 if not)
    - Plane 26 for squares attacked by opponent bishops (1 if attacked, 0 if not)
    - Plane 27 for squares attacked by opponent knights (1 if attacked, 0 if not)
    - Plane 28 for squares attacked by opponent pawns (1 if attacked, 0 if not)
    """

    # Initialize tensor
    tensorboard = torch.zeros(29, 8, 8)

    # Mirror the board if black is playing to make the model color-invartiant
    if mirror_if_black and chessboard.turn == chess.BLACK:
        chessboard = chessboard.mirror()

    # Define the piece type and color mappings
    map_piece_to_plane = {
        (chess.KING, chess.WHITE): 0,
        (chess.QUEEN, chess.WHITE): 1,
        (chess.ROOK, chess.WHITE): 2,
        (chess.BISHOP, chess.WHITE): 3,
        (chess.KNIGHT, chess.WHITE): 4,
        (chess.PAWN, chess.WHITE): 5,
        (chess.KING, chess.BLACK): 6,
        (chess.QUEEN, chess.BLACK): 7,
        (chess.ROOK, chess.BLACK): 8,
        (chess.BISHOP, chess.BLACK): 9,
        (chess.KNIGHT, chess.BLACK): 10,
        (chess.PAWN, chess.BLACK): 11
    }

    # Get the piece map dictionary
    piece_map: dict[chess.Square, chess.Piece] = chessboard.piece_map()

    ## P0-P11: PIECES POSITIONS

    # Iterate over each square on the board
    for square in chess.SQUARES:
        piece = piece_map.get(square)
        if piece:
            plane = map_piece_to_plane.get((piece.piece_type, piece.color))
            if plane is not None:
                row, col = divmod(square, 8)
                tensorboard[plane, row, col] = 1
    

    ## P12-P15: CASTLING RIGHTS

    # P12: player castling rights king side
    if chessboard.has_kingside_castling_rights(chess.WHITE):
        tensorboard[12, :, :] = 1
    
    # P13: player castling rights queen side
    if chessboard.has_queenside_castling_rights(chess.WHITE):
        tensorboard[13, :, :] = 1
    
    # P14: opponent castling rights king side
    if chessboard.has_kingside_castling_rights(chess.BLACK):
        tensorboard[14, :, :] = 1
    
    # P15: opponent castling rights queen side
    if chessboard.has_queenside_castling_rights(chess.BLACK):
        tensorboard[15, :, :] = 1


    ## P16: EN PASSANT SQUARE
    if chessboard.ep_square is not None:
        row, col = divmod(chessboard.ep_square, 8)
        tensorboard[16, row, col] = 1


    ## P17-P28: ATTACKED SQUARES

    # Iterate over each square on the board
    for square in chess.SQUARES:
        attackers = chessboard.attackers(chess.WHITE, square)
        for attacker in attackers:
            piece = piece_map.get(attacker)
            if piece:
                piece_type = piece.piece_type
                row, col = divmod(square, 8)
                if piece_type == chess.KING:
                    tensorboard[17, row, col] = 1
                elif piece_type == chess.QUEEN:
                    tensorboard[18, row, col] = 1
                elif piece_type == chess.ROOK:
                    tensorboard[19, row, col] = 1
                elif piece_type == chess.BISHOP:
                    tensorboard[20, row, col] = 1
                elif piece_type == chess.KNIGHT:
                    tensorboard[21, row, col] = 1
                elif piece_type == chess.PAWN:
                    tensorboard[22, row, col] = 1

        attackers = chessboard.attackers(chess.BLACK, square)
        for attacker in attackers:
            piece = piece_map.get(attacker)
            if piece:
                piece_type = piece.piece_type
                row, col = divmod(square, 8)
                if piece_type == chess.KING:
                    tensorboard[23, row, col] = 1
                elif piece_type == chess.QUEEN:
                    tensorboard[24, row, col] = 1
                elif piece_type == chess.ROOK:
                    tensorboard[25, row, col] = 1
                elif piece_type == chess.BISHOP:
                    tensorboard[26, row, col] = 1
                elif piece_type == chess.KNIGHT:
                    tensorboard[27, row, col] = 1
                elif piece_type == chess.PAWN:
                    tensorboard[28, row, col] = 1

    return tensorboard

DICT_1858_MOVES_INDEX_TO_UCI:dict[int,str] = {
    0: "a1b1", 1: "a1c1", 2: "a1d1", 3: "a1e1", 4: "a1f1", 5: "a1g1", 6: "a1h1", 7: "a1a2",
    8: "a1b2", 9: "a1c2", 10: "a1a3", 11: "a1b3", 12: "a1c3", 13: "a1a4", 14: "a1d4", 15: "a1a5",
    16: "a1e5", 17: "a1a6", 18: "a1f6", 19: "a1a7", 20: "a1g7", 21: "a1a8", 22: "a1h8", 23: "b1a1",
    24: "b1c1", 25: "b1d1", 26: "b1e1", 27: "b1f1", 28: "b1g1", 29: "b1h1", 30: "b1a2", 31: "b1b2",
    32: "b1c2", 33: "b1d2", 34: "b1a3", 35: "b1b3", 36: "b1c3", 37: "b1d3", 38: "b1b4", 39: "b1e4",
    40: "b1b5", 41: "b1f5", 42: "b1b6", 43: "b1g6", 44: "b1b7", 45: "b1h7", 46: "b1b8", 47: "c1a1",
    48: "c1b1", 49: "c1d1", 50: "c1e1", 51: "c1f1", 52: "c1g1", 53: "c1h1", 54: "c1a2", 55: "c1b2",
    56: "c1c2", 57: "c1d2", 58: "c1e2", 59: "c1a3", 60: "c1b3", 61: "c1c3", 62: "c1d3", 63: "c1e3",
    64: "c1c4", 65: "c1f4", 66: "c1c5", 67: "c1g5", 68: "c1c6", 69: "c1h6", 70: "c1c7", 71: "c1c8",
    72: "d1a1", 73: "d1b1", 74: "d1c1", 75: "d1e1", 76: "d1f1", 77: "d1g1", 78: "d1h1", 79: "d1b2",
    80: "d1c2", 81: "d1d2", 82: "d1e2", 83: "d1f2", 84: "d1b3", 85: "d1c3", 86: "d1d3", 87: "d1e3",
    88: "d1f3", 89: "d1a4", 90: "d1d4", 91: "d1g4", 92: "d1d5", 93: "d1h5", 94: "d1d6", 95: "d1d7",
    96: "d1d8", 97: "e1a1", 98: "e1b1", 99: "e1c1", 100: "e1d1", 101: "e1f1", 102: "e1g1", 103: "e1h1",
    104: "e1c2", 105: "e1d2", 106: "e1e2", 107: "e1f2", 108: "e1g2", 109: "e1c3", 110: "e1d3", 111: "e1e3",
    112: "e1f3", 113: "e1g3", 114: "e1b4", 115: "e1e4", 116: "e1h4", 117: "e1a5", 118: "e1e5", 119: "e1e6",
    120: "e1e7", 121: "e1e8", 122: "f1a1", 123: "f1b1", 124: "f1c1", 125: "f1d1", 126: "f1e1", 127: "f1g1",
    128: "f1h1", 129: "f1d2", 130: "f1e2", 131: "f1f2", 132: "f1g2", 133: "f1h2", 134: "f1d3", 135: "f1e3",
    136: "f1f3", 137: "f1g3", 138: "f1h3", 139: "f1c4", 140: "f1f4", 141: "f1b5", 142: "f1f5", 143: "f1a6",
    144: "f1f6", 145: "f1f7", 146: "f1f8", 147: "g1a1", 148: "g1b1", 149: "g1c1", 150: "g1d1", 151: "g1e1",
    152: "g1f1", 153: "g1h1", 154: "g1e2", 155: "g1f2", 156: "g1g2", 157: "g1h2", 158: "g1e3", 159: "g1f3",
    160: "g1g3", 161: "g1h3", 162: "g1d4", 163: "g1g4", 164: "g1c5", 165: "g1g5", 166: "g1b6", 167: "g1g6",
    168: "g1a7", 169: "g1g7", 170: "g1g8", 171: "h1a1", 172: "h1b1", 173: "h1c1", 174: "h1d1", 175: "h1e1",
    176: "h1f1", 177: "h1g1", 178: "h1f2", 179: "h1g2", 180: "h1h2", 181: "h1f3", 182: "h1g3", 183: "h1h3",
    184: "h1e4", 185: "h1h4", 186: "h1d5", 187: "h1h5", 188: "h1c6", 189: "h1h6", 190: "h1b7", 191: "h1h7",
    192: "h1a8", 193: "h1h8", 194: "a2a1", 195: "a2b1", 196: "a2c1", 197: "a2b2", 198: "a2c2", 199: "a2d2",
    200: "a2e2", 201: "a2f2", 202: "a2g2", 203: "a2h2", 204: "a2a3", 205: "a2b3", 206: "a2c3", 207: "a2a4",
    208: "a2b4", 209: "a2c4", 210: "a2a5", 211: "a2d5", 212: "a2a6", 213: "a2e6", 214: "a2a7", 215: "a2f7",
    216: "a2a8", 217: "a2g8", 218: "b2a1", 219: "b2b1", 220: "b2c1", 221: "b2d1", 222: "b2a2", 223: "b2c2",
    224: "b2d2", 225: "b2e2", 226: "b2f2", 227: "b2g2", 228: "b2h2", 229: "b2a3", 230: "b2b3", 231: "b2c3",
    232: "b2d3", 233: "b2a4", 234: "b2b4", 235: "b2c4", 236: "b2d4", 237: "b2b5", 238: "b2e5", 239: "b2b6",
    240: "b2f6", 241: "b2b7", 242: "b2g7", 243: "b2b8", 244: "b2h8", 245: "c2a1", 246: "c2b1", 247: "c2c1",
    248: "c2d1", 249: "c2e1", 250: "c2a2", 251: "c2b2", 252: "c2d2", 253: "c2e2", 254: "c2f2", 255: "c2g2",
    256: "c2h2", 257: "c2a3", 258: "c2b3", 259: "c2c3", 260: "c2d3", 261: "c2e3", 262: "c2a4", 263: "c2b4",
    264: "c2c4", 265: "c2d4", 266: "c2e4", 267: "c2c5", 268: "c2f5", 269: "c2c6", 270: "c2g6", 271: "c2c7",
    272: "c2h7", 273: "c2c8", 274: "d2b1", 275: "d2c1", 276: "d2d1", 277: "d2e1", 278: "d2f1", 279: "d2a2",
    280: "d2b2", 281: "d2c2", 282: "d2e2", 283: "d2f2", 284: "d2g2", 285: "d2h2", 286: "d2b3", 287: "d2c3",
    288: "d2d3", 289: "d2e3", 290: "d2f3", 291: "d2b4", 292: "d2c4", 293: "d2d4", 294: "d2e4", 295: "d2f4",
    296: "d2a5", 297: "d2d5", 298: "d2g5", 299: "d2d6", 300: "d2h6", 301: "d2d7", 302: "d2d8", 303: "e2c1",
    304: "e2d1", 305: "e2e1", 306: "e2f1", 307: "e2g1", 308: "e2a2", 309: "e2b2", 310: "e2c2", 311: "e2d2",
    312: "e2f2", 313: "e2g2", 314: "e2h2", 315: "e2c3", 316: "e2d3", 317: "e2e3", 318: "e2f3", 319: "e2g3",
    320: "e2c4", 321: "e2d4", 322: "e2e4", 323: "e2f4", 324: "e2g4", 325: "e2b5", 326: "e2e5", 327: "e2h5",
    328: "e2a6", 329: "e2e6", 330: "e2e7", 331: "e2e8", 332: "f2d1", 333: "f2e1", 334: "f2f1", 335: "f2g1",
    336: "f2h1", 337: "f2a2", 338: "f2b2", 339: "f2c2", 340: "f2d2", 341: "f2e2", 342: "f2g2", 343: "f2h2",
    344: "f2d3", 345: "f2e3", 346: "f2f3", 347: "f2g3", 348: "f2h3", 349: "f2d4", 350: "f2e4", 351: "f2f4",
    352: "f2g4", 353: "f2h4", 354: "f2c5", 355: "f2f5", 356: "f2b6", 357: "f2f6", 358: "f2a7", 359: "f2f7",
    360: "f2f8", 361: "g2e1", 362: "g2f1", 363: "g2g1", 364: "g2h1", 365: "g2a2", 366: "g2b2", 367: "g2c2",
    368: "g2d2", 369: "g2e2", 370: "g2f2", 371: "g2h2", 372: "g2e3", 373: "g2f3", 374: "g2g3", 375: "g2h3",
    376: "g2e4", 377: "g2f4", 378: "g2g4", 379: "g2h4", 380: "g2d5", 381: "g2g5", 382: "g2c6", 383: "g2g6",
    384: "g2b7", 385: "g2g7", 386: "g2a8", 387: "g2g8", 388: "h2f1", 389: "h2g1", 390: "h2h1", 391: "h2a2",
    392: "h2b2", 393: "h2c2", 394: "h2d2", 395: "h2e2", 396: "h2f2", 397: "h2g2", 398: "h2f3", 399: "h2g3",
    400: "h2h3", 401: "h2f4", 402: "h2g4", 403: "h2h4", 404: "h2e5", 405: "h2h5", 406: "h2d6", 407: "h2h6",
    408: "h2c7", 409: "h2h7", 410: "h2b8", 411: "h2h8", 412: "a3a1", 413: "a3b1", 414: "a3c1", 415: "a3a2",
    416: "a3b2", 417: "a3c2", 418: "a3b3", 419: "a3c3", 420: "a3d3", 421: "a3e3", 422: "a3f3", 423: "a3g3",
    424: "a3h3", 425: "a3a4", 426: "a3b4", 427: "a3c4", 428: "a3a5", 429: "a3b5", 430: "a3c5", 431: "a3a6",
    432: "a3d6", 433: "a3a7", 434: "a3e7", 435: "a3a8", 436: "a3f8", 437: "b3a1", 438: "b3b1", 439: "b3c1",
    440: "b3d1", 441: "b3a2", 442: "b3b2", 443: "b3c2", 444: "b3d2", 445: "b3a3", 446: "b3c3", 447: "b3d3",
    448: "b3e3", 449: "b3f3", 450: "b3g3", 451: "b3h3", 452: "b3a4", 453: "b3b4", 454: "b3c4", 455: "b3d4",
    456: "b3a5", 457: "b3b5", 458: "b3c5", 459: "b3d5", 460: "b3b6", 461: "b3e6", 462: "b3b7", 463: "b3f7",
    464: "b3b8", 465: "b3g8", 466: "c3a1", 467: "c3b1", 468: "c3c1", 469: "c3d1", 470: "c3e1", 471: "c3a2",
    472: "c3b2", 473: "c3c2", 474: "c3d2", 475: "c3e2", 476: "c3a3", 477: "c3b3", 478: "c3d3", 479: "c3e3",
    480: "c3f3", 481: "c3g3", 482: "c3h3", 483: "c3a4", 484: "c3b4", 485: "c3c4", 486: "c3d4", 487: "c3e4",
    488: "c3a5", 489: "c3b5", 490: "c3c5", 491: "c3d5", 492: "c3e5", 493: "c3c6", 494: "c3f6", 495: "c3c7",
    496: "c3g7", 497: "c3c8", 498: "c3h8", 499: "d3b1", 500: "d3c1", 501: "d3d1", 502: "d3e1", 503: "d3f1",
    504: "d3b2", 505: "d3c2", 506: "d3d2", 507: "d3e2", 508: "d3f2", 509: "d3a3", 510: "d3b3", 511: "d3c3",
    512: "d3e3", 513: "d3f3", 514: "d3g3", 515: "d3h3", 516: "d3b4", 517: "d3c4", 518: "d3d4", 519: "d3e4",
    520: "d3f4", 521: "d3b5", 522: "d3c5", 523: "d3d5", 524: "d3e5", 525: "d3f5", 526: "d3a6", 527: "d3d6",
    528: "d3g6", 529: "d3d7", 530: "d3h7", 531: "d3d8", 532: "e3c1", 533: "e3d1", 534: "e3e1", 535: "e3f1",
    536: "e3g1", 537: "e3c2", 538: "e3d2", 539: "e3e2", 540: "e3f2", 541: "e3g2", 542: "e3a3", 543: "e3b3",
    544: "e3c3", 545: "e3d3", 546: "e3f3", 547: "e3g3", 548: "e3h3", 549: "e3c4", 550: "e3d4", 551: "e3e4",
    552: "e3f4", 553: "e3g4", 554: "e3c5", 555: "e3d5", 556: "e3e5", 557: "e3f5", 558: "e3g5", 559: "e3b6",
    560: "e3e6", 561: "e3h6", 562: "e3a7", 563: "e3e7", 564: "e3e8", 565: "f3d1", 566: "f3e1", 567: "f3f1",
    568: "f3g1", 569: "f3h1", 570: "f3d2", 571: "f3e2", 572: "f3f2", 573: "f3g2", 574: "f3h2", 575: "f3a3",
    576: "f3b3", 577: "f3c3", 578: "f3d3", 579: "f3e3", 580: "f3g3", 581: "f3h3", 582: "f3d4", 583: "f3e4",
    584: "f3f4", 585: "f3g4", 586: "f3h4", 587: "f3d5", 588: "f3e5", 589: "f3f5", 590: "f3g5", 591: "f3h5",
    592: "f3c6", 593: "f3f6", 594: "f3b7", 595: "f3f7", 596: "f3a8", 597: "f3f8", 598: "g3e1", 599: "g3f1",
    600: "g3g1", 601: "g3h1", 602: "g3e2", 603: "g3f2", 604: "g3g2", 605: "g3h2", 606: "g3a3", 607: "g3b3",
    608: "g3c3", 609: "g3d3", 610: "g3e3", 611: "g3f3", 612: "g3h3", 613: "g3e4", 614: "g3f4", 615: "g3g4",
    616: "g3h4", 617: "g3e5", 618: "g3f5", 619: "g3g5", 620: "g3h5", 621: "g3d6", 622: "g3g6", 623: "g3c7",
    624: "g3g7", 625: "g3b8", 626: "g3g8", 627: "h3f1", 628: "h3g1", 629: "h3h1", 630: "h3f2", 631: "h3g2",
    632: "h3h2", 633: "h3a3", 634: "h3b3", 635: "h3c3", 636: "h3d3", 637: "h3e3", 638: "h3f3", 639: "h3g3",
    640: "h3f4", 641: "h3g4", 642: "h3h4", 643: "h3f5", 644: "h3g5", 645: "h3h5", 646: "h3e6", 647: "h3h6",
    648: "h3d7", 649: "h3h7", 650: "h3c8", 651: "h3h8", 652: "a4a1", 653: "a4d1", 654: "a4a2", 655: "a4b2",
    656: "a4c2", 657: "a4a3", 658: "a4b3", 659: "a4c3", 660: "a4b4", 661: "a4c4", 662: "a4d4", 663: "a4e4",
    664: "a4f4", 665: "a4g4", 666: "a4h4", 667: "a4a5", 668: "a4b5", 669: "a4c5", 670: "a4a6", 671: "a4b6",
    672: "a4c6", 673: "a4a7", 674: "a4d7", 675: "a4a8", 676: "a4e8", 677: "b4b1", 678: "b4e1", 679: "b4a2",
    680: "b4b2", 681: "b4c2", 682: "b4d2", 683: "b4a3", 684: "b4b3", 685: "b4c3", 686: "b4d3", 687: "b4a4",
    688: "b4c4", 689: "b4d4", 690: "b4e4", 691: "b4f4", 692: "b4g4", 693: "b4h4", 694: "b4a5", 695: "b4b5",
    696: "b4c5", 697: "b4d5", 698: "b4a6", 699: "b4b6", 700: "b4c6", 701: "b4d6", 702: "b4b7", 703: "b4e7",
    704: "b4b8", 705: "b4f8", 706: "c4c1", 707: "c4f1", 708: "c4a2", 709: "c4b2", 710: "c4c2", 711: "c4d2",
    712: "c4e2", 713: "c4a3", 714: "c4b3", 715: "c4c3", 716: "c4d3", 717: "c4e3", 718: "c4a4", 719: "c4b4",
    720: "c4d4", 721: "c4e4", 722: "c4f4", 723: "c4g4", 724: "c4h4", 725: "c4a5", 726: "c4b5", 727: "c4c5",
    728: "c4d5", 729: "c4e5", 730: "c4a6", 731: "c4b6", 732: "c4c6", 733: "c4d6", 734: "c4e6", 735: "c4c7",
    736: "c4f7", 737: "c4c8", 738: "c4g8", 739: "d4a1", 740: "d4d1", 741: "d4g1", 742: "d4b2", 743: "d4c2",
    744: "d4d2", 745: "d4e2", 746: "d4f2", 747: "d4b3", 748: "d4c3", 749: "d4d3", 750: "d4e3", 751: "d4f3",
    752: "d4a4", 753: "d4b4", 754: "d4c4", 755: "d4e4", 756: "d4f4", 757: "d4g4", 758: "d4h4", 759: "d4b5",
    760: "d4c5", 761: "d4d5", 762: "d4e5", 763: "d4f5", 764: "d4b6", 765: "d4c6", 766: "d4d6", 767: "d4e6",
    768: "d4f6", 769: "d4a7", 770: "d4d7", 771: "d4g7", 772: "d4d8", 773: "d4h8", 774: "e4b1", 775: "e4e1",
    776: "e4h1", 777: "e4c2", 778: "e4d2", 779: "e4e2", 780: "e4f2", 781: "e4g2", 782: "e4c3", 783: "e4d3",
    784: "e4e3", 785: "e4f3", 786: "e4g3", 787: "e4a4", 788: "e4b4", 789: "e4c4", 790: "e4d4", 791: "e4f4",
    792: "e4g4", 793: "e4h4", 794: "e4c5", 795: "e4d5", 796: "e4e5", 797: "e4f5", 798: "e4g5", 799: "e4c6",
    800: "e4d6", 801: "e4e6", 802: "e4f6", 803: "e4g6", 804: "e4b7", 805: "e4e7", 806: "e4h7", 807: "e4a8",
    808: "e4e8", 809: "f4c1", 810: "f4f1", 811: "f4d2", 812: "f4e2", 813: "f4f2", 814: "f4g2", 815: "f4h2",
    816: "f4d3", 817: "f4e3", 818: "f4f3", 819: "f4g3", 820: "f4h3", 821: "f4a4", 822: "f4b4", 823: "f4c4",
    824: "f4d4", 825: "f4e4", 826: "f4g4", 827: "f4h4", 828: "f4d5", 829: "f4e5", 830: "f4f5", 831: "f4g5",
    832: "f4h5", 833: "f4d6", 834: "f4e6", 835: "f4f6", 836: "f4g6", 837: "f4h6", 838: "f4c7", 839: "f4f7",
    840: "f4b8", 841: "f4f8", 842: "g4d1", 843: "g4g1", 844: "g4e2", 845: "g4f2", 846: "g4g2", 847: "g4h2",
    848: "g4e3", 849: "g4f3", 850: "g4g3", 851: "g4h3", 852: "g4a4", 853: "g4b4", 854: "g4c4", 855: "g4d4",
    856: "g4e4", 857: "g4f4", 858: "g4h4", 859: "g4e5", 860: "g4f5", 861: "g4g5", 862: "g4h5", 863: "g4e6",
    864: "g4f6", 865: "g4g6", 866: "g4h6", 867: "g4d7", 868: "g4g7", 869: "g4c8", 870: "g4g8", 871: "h4e1",
    872: "h4h1", 873: "h4f2", 874: "h4g2", 875: "h4h2", 876: "h4f3", 877: "h4g3", 878: "h4h3", 879: "h4a4",
    880: "h4b4", 881: "h4c4", 882: "h4d4", 883: "h4e4", 884: "h4f4", 885: "h4g4", 886: "h4f5", 887: "h4g5",
    888: "h4h5", 889: "h4f6", 890: "h4g6", 891: "h4h6", 892: "h4e7", 893: "h4h7", 894: "h4d8", 895: "h4h8",
    896: "a5a1", 897: "a5e1", 898: "a5a2", 899: "a5d2", 900: "a5a3", 901: "a5b3", 902: "a5c3", 903: "a5a4",
    904: "a5b4", 905: "a5c4", 906: "a5b5", 907: "a5c5", 908: "a5d5", 909: "a5e5", 910: "a5f5", 911: "a5g5",
    912: "a5h5", 913: "a5a6", 914: "a5b6", 915: "a5c6", 916: "a5a7", 917: "a5b7", 918: "a5c7", 919: "a5a8",
    920: "a5d8", 921: "b5b1", 922: "b5f1", 923: "b5b2", 924: "b5e2", 925: "b5a3", 926: "b5b3", 927: "b5c3",
    928: "b5d3", 929: "b5a4", 930: "b5b4", 931: "b5c4", 932: "b5d4", 933: "b5a5", 934: "b5c5", 935: "b5d5",
    936: "b5e5", 937: "b5f5", 938: "b5g5", 939: "b5h5", 940: "b5a6", 941: "b5b6", 942: "b5c6", 943: "b5d6",
    944: "b5a7", 945: "b5b7", 946: "b5c7", 947: "b5d7", 948: "b5b8", 949: "b5e8", 950: "c5c1", 951: "c5g1",
    952: "c5c2", 953: "c5f2", 954: "c5a3", 955: "c5b3", 956: "c5c3", 957: "c5d3", 958: "c5e3", 959: "c5a4",
    960: "c5b4", 961: "c5c4", 962: "c5d4", 963: "c5e4", 964: "c5a5", 965: "c5b5", 966: "c5d5", 967: "c5e5",
    968: "c5f5", 969: "c5g5", 970: "c5h5", 971: "c5a6", 972: "c5b6", 973: "c5c6", 974: "c5d6", 975: "c5e6",
    976: "c5a7", 977: "c5b7", 978: "c5c7", 979: "c5d7", 980: "c5e7", 981: "c5c8", 982: "c5f8", 983: "d5d1",
    984: "d5h1", 985: "d5a2", 986: "d5d2", 987: "d5g2", 988: "d5b3", 989: "d5c3", 990: "d5d3", 991: "d5e3",
    992: "d5f3", 993: "d5b4", 994: "d5c4", 995: "d5d4", 996: "d5e4", 997: "d5f4", 998: "d5a5", 999: "d5b5",
    1000: "d5c5", 1001: "d5e5", 1002: "d5f5", 1003: "d5g5", 1004: "d5h5", 1005: "d5b6", 1006: "d5c6", 1007: "d5d6",
    1008: "d5e6", 1009: "d5f6", 1010: "d5b7", 1011: "d5c7", 1012: "d5d7", 1013: "d5e7", 1014: "d5f7", 1015: "d5a8",
    1016: "d5d8", 1017: "d5g8", 1018: "e5a1", 1019: "e5e1", 1020: "e5b2", 1021: "e5e2", 1022: "e5h2", 1023: "e5c3",
    1024: "e5d3", 1025: "e5e3", 1026: "e5f3", 1027: "e5g3", 1028: "e5c4", 1029: "e5d4", 1030: "e5e4", 1031: "e5f4",
    1032: "e5g4", 1033: "e5a5", 1034: "e5b5", 1035: "e5c5", 1036: "e5d5", 1037: "e5f5", 1038: "e5g5", 1039: "e5h5",
    1040: "e5c6", 1041: "e5d6", 1042: "e5e6", 1043: "e5f6", 1044: "e5g6", 1045: "e5c7", 1046: "e5d7", 1047: "e5e7",
    1048: "e5f7", 1049: "e5g7", 1050: "e5b8", 1051: "e5e8", 1052: "e5h8", 1053: "f5b1", 1054: "f5f1", 1055: "f5c2",
    1056: "f5f2", 1057: "f5d3", 1058: "f5e3", 1059: "f5f3", 1060: "f5g3", 1061: "f5h3", 1062: "f5d4", 1063: "f5e4",
    1064: "f5f4", 1065: "f5g4", 1066: "f5h4", 1067: "f5a5", 1068: "f5b5", 1069: "f5c5", 1070: "f5d5", 1071: "f5e5",
    1072: "f5g5", 1073: "f5h5", 1074: "f5d6", 1075: "f5e6", 1076: "f5f6", 1077: "f5g6", 1078: "f5h6", 1079: "f5d7",
    1080: "f5e7", 1081: "f5f7", 1082: "f5g7", 1083: "f5h7", 1084: "f5c8", 1085: "f5f8", 1086: "g5c1", 1087: "g5g1",
    1088: "g5d2", 1089: "g5g2", 1090: "g5e3", 1091: "g5f3", 1092: "g5g3", 1093: "g5h3", 1094: "g5e4", 1095: "g5f4",
    1096: "g5g4", 1097: "g5h4", 1098: "g5a5", 1099: "g5b5", 1100: "g5c5", 1101: "g5d5", 1102: "g5e5", 1103: "g5f5",
    1104: "g5h5", 1105: "g5e6", 1106: "g5f6", 1107: "g5g6", 1108: "g5h6", 1109: "g5e7", 1110: "g5f7", 1111: "g5g7",
    1112: "g5h7", 1113: "g5d8", 1114: "g5g8", 1115: "h5d1", 1116: "h5h1", 1117: "h5e2", 1118: "h5h2", 1119: "h5f3",
    1120: "h5g3", 1121: "h5h3", 1122: "h5f4", 1123: "h5g4", 1124: "h5h4", 1125: "h5a5", 1126: "h5b5", 1127: "h5c5",
    1128: "h5d5", 1129: "h5e5", 1130: "h5f5", 1131: "h5g5", 1132: "h5f6", 1133: "h5g6", 1134: "h5h6", 1135: "h5f7",
    1136: "h5g7", 1137: "h5h7", 1138: "h5e8", 1139: "h5h8", 1140: "a6a1", 1141: "a6f1", 1142: "a6a2", 1143: "a6e2",
    1144: "a6a3", 1145: "a6d3", 1146: "a6a4", 1147: "a6b4", 1148: "a6c4", 1149: "a6a5", 1150: "a6b5", 1151: "a6c5",
    1152: "a6b6", 1153: "a6c6", 1154: "a6d6", 1155: "a6e6", 1156: "a6f6", 1157: "a6g6", 1158: "a6h6", 1159: "a6a7",
    1160: "a6b7", 1161: "a6c7", 1162: "a6a8", 1163: "a6b8", 1164: "a6c8", 1165: "b6b1", 1166: "b6g1", 1167: "b6b2",
    1168: "b6f2", 1169: "b6b3", 1170: "b6e3", 1171: "b6a4", 1172: "b6b4", 1173: "b6c4", 1174: "b6d4", 1175: "b6a5",
    1176: "b6b5", 1177: "b6c5", 1178: "b6d5", 1179: "b6a6", 1180: "b6c6", 1181: "b6d6", 1182: "b6e6", 1183: "b6f6",
    1184: "b6g6", 1185: "b6h6", 1186: "b6a7", 1187: "b6b7", 1188: "b6c7", 1189: "b6d7", 1190: "b6a8", 1191: "b6b8",
    1192: "b6c8", 1193: "b6d8", 1194: "c6c1", 1195: "c6h1", 1196: "c6c2", 1197: "c6g2", 1198: "c6c3", 1199: "c6f3",
    1200: "c6a4", 1201: "c6b4", 1202: "c6c4", 1203: "c6d4", 1204: "c6e4", 1205: "c6a5", 1206: "c6b5", 1207: "c6c5",
    1208: "c6d5", 1209: "c6e5", 1210: "c6a6", 1211: "c6b6", 1212: "c6d6", 1213: "c6e6", 1214: "c6f6", 1215: "c6g6",
    1216: "c6h6", 1217: "c6a7", 1218: "c6b7", 1219: "c6c7", 1220: "c6d7", 1221: "c6e7", 1222: "c6a8", 1223: "c6b8",
    1224: "c6c8", 1225: "c6d8", 1226: "c6e8", 1227: "d6d1", 1228: "d6d2", 1229: "d6h2", 1230: "d6a3", 1231: "d6d3",
    1232: "d6g3", 1233: "d6b4", 1234: "d6c4", 1235: "d6d4", 1236: "d6e4", 1237: "d6f4", 1238: "d6b5", 1239: "d6c5",
    1240: "d6d5", 1241: "d6e5", 1242: "d6f5", 1243: "d6a6", 1244: "d6b6", 1245: "d6c6", 1246: "d6e6", 1247: "d6f6",
    1248: "d6g6", 1249: "d6h6", 1250: "d6b7", 1251: "d6c7", 1252: "d6d7", 1253: "d6e7", 1254: "d6f7", 1255: "d6b8",
    1256: "d6c8", 1257: "d6d8", 1258: "d6e8", 1259: "d6f8", 1260: "e6e1", 1261: "e6a2", 1262: "e6e2", 1263: "e6b3",
    1264: "e6e3", 1265: "e6h3", 1266: "e6c4", 1267: "e6d4", 1268: "e6e4", 1269: "e6f4", 1270: "e6g4", 1271: "e6c5",
    1272: "e6d5", 1273: "e6e5", 1274: "e6f5", 1275: "e6g5", 1276: "e6a6", 1277: "e6b6", 1278: "e6c6", 1279: "e6d6",
    1280: "e6f6", 1281: "e6g6", 1282: "e6h6", 1283: "e6c7", 1284: "e6d7", 1285: "e6e7", 1286: "e6f7", 1287: "e6g7",
    1288: "e6c8", 1289: "e6d8", 1290: "e6e8", 1291: "e6f8", 1292: "e6g8", 1293: "f6a1", 1294: "f6f1", 1295: "f6b2",
    1296: "f6f2", 1297: "f6c3", 1298: "f6f3", 1299: "f6d4", 1300: "f6e4", 1301: "f6f4", 1302: "f6g4", 1303: "f6h4",
    1304: "f6d5", 1305: "f6e5", 1306: "f6f5", 1307: "f6g5", 1308: "f6h5", 1309: "f6a6", 1310: "f6b6", 1311: "f6c6",
    1312: "f6d6", 1313: "f6e6", 1314: "f6g6", 1315: "f6h6", 1316: "f6d7", 1317: "f6e7", 1318: "f6f7", 1319: "f6g7",
    1320: "f6h7", 1321: "f6d8", 1322: "f6e8", 1323: "f6f8", 1324: "f6g8", 1325: "f6h8", 1326: "g6b1", 1327: "g6g1",
    1328: "g6c2", 1329: "g6g2", 1330: "g6d3", 1331: "g6g3", 1332: "g6e4", 1333: "g6f4", 1334: "g6g4", 1335: "g6h4",
    1336: "g6e5", 1337: "g6f5", 1338: "g6g5", 1339: "g6h5", 1340: "g6a6", 1341: "g6b6", 1342: "g6c6", 1343: "g6d6",
    1344: "g6e6", 1345: "g6f6", 1346: "g6h6", 1347: "g6e7", 1348: "g6f7", 1349: "g6g7", 1350: "g6h7", 1351: "g6e8",
    1352: "g6f8", 1353: "g6g8", 1354: "g6h8", 1355: "h6c1", 1356: "h6h1", 1357: "h6d2", 1358: "h6h2", 1359: "h6e3",
    1360: "h6h3", 1361: "h6f4", 1362: "h6g4", 1363: "h6h4", 1364: "h6f5", 1365: "h6g5", 1366: "h6h5", 1367: "h6a6",
    1368: "h6b6", 1369: "h6c6", 1370: "h6d6", 1371: "h6e6", 1372: "h6f6", 1373: "h6g6", 1374: "h6f7", 1375: "h6g7",
    1376: "h6h7", 1377: "h6f8", 1378: "h6g8", 1379: "h6h8", 1380: "a7a1", 1381: "a7g1", 1382: "a7a2", 1383: "a7f2",
    1384: "a7a3", 1385: "a7e3", 1386: "a7a4", 1387: "a7d4", 1388: "a7a5", 1389: "a7b5", 1390: "a7c5", 1391: "a7a6",
    1392: "a7b6", 1393: "a7c6", 1394: "a7b7", 1395: "a7c7", 1396: "a7d7", 1397: "a7e7", 1398: "a7f7", 1399: "a7g7",
    1400: "a7h7", 1401: "a7a8", 1402: "a7b8", 1403: "a7c8", 1404: "b7b1", 1405: "b7h1", 1406: "b7b2", 1407: "b7g2",
    1408: "b7b3", 1409: "b7f3", 1410: "b7b4", 1411: "b7e4", 1412: "b7a5", 1413: "b7b5", 1414: "b7c5", 1415: "b7d5",
    1416: "b7a6", 1417: "b7b6", 1418: "b7c6", 1419: "b7d6", 1420: "b7a7", 1421: "b7c7", 1422: "b7d7", 1423: "b7e7",
    1424: "b7f7", 1425: "b7g7", 1426: "b7h7", 1427: "b7a8", 1428: "b7b8", 1429: "b7c8", 1430: "b7d8", 1431: "c7c1",
    1432: "c7c2", 1433: "c7h2", 1434: "c7c3", 1435: "c7g3", 1436: "c7c4", 1437: "c7f4", 1438: "c7a5", 1439: "c7b5",
    1440: "c7c5", 1441: "c7d5", 1442: "c7e5", 1443: "c7a6", 1444: "c7b6", 1445: "c7c6", 1446: "c7d6", 1447: "c7e6",
    1448: "c7a7", 1449: "c7b7", 1450: "c7d7", 1451: "c7e7", 1452: "c7f7", 1453: "c7g7", 1454: "c7h7", 1455: "c7a8",
    1456: "c7b8", 1457: "c7c8", 1458: "c7d8", 1459: "c7e8", 1460: "d7d1", 1461: "d7d2", 1462: "d7d3", 1463: "d7h3",
    1464: "d7a4", 1465: "d7d4", 1466: "d7g4", 1467: "d7b5", 1468: "d7c5", 1469: "d7d5", 1470: "d7e5", 1471: "d7f5",
    1472: "d7b6", 1473: "d7c6", 1474: "d7d6", 1475: "d7e6", 1476: "d7f6", 1477: "d7a7", 1478: "d7b7", 1479: "d7c7",
    1480: "d7e7", 1481: "d7f7", 1482: "d7g7", 1483: "d7h7", 1484: "d7b8", 1485: "d7c8", 1486: "d7d8", 1487: "d7e8",
    1488: "d7f8", 1489: "e7e1", 1490: "e7e2", 1491: "e7a3", 1492: "e7e3", 1493: "e7b4", 1494: "e7e4", 1495: "e7h4",
    1496: "e7c5", 1497: "e7d5", 1498: "e7e5", 1499: "e7f5", 1500: "e7g5", 1501: "e7c6", 1502: "e7d6", 1503: "e7e6",
    1504: "e7f6", 1505: "e7g6", 1506: "e7a7", 1507: "e7b7", 1508: "e7c7", 1509: "e7d7", 1510: "e7f7", 1511: "e7g7",
    1512: "e7h7", 1513: "e7c8", 1514: "e7d8", 1515: "e7e8", 1516: "e7f8", 1517: "e7g8", 1518: "f7f1", 1519: "f7a2",
    1520: "f7f2", 1521: "f7b3", 1522: "f7f3", 1523: "f7c4", 1524: "f7f4", 1525: "f7d5", 1526: "f7e5", 1527: "f7f5",
    1528: "f7g5", 1529: "f7h5", 1530: "f7d6", 1531: "f7e6", 1532: "f7f6", 1533: "f7g6", 1534: "f7h6", 1535: "f7a7",
    1536: "f7b7", 1537: "f7c7", 1538: "f7d7", 1539: "f7e7", 1540: "f7g7", 1541: "f7h7", 1542: "f7d8", 1543: "f7e8",
    1544: "f7f8", 1545: "f7g8", 1546: "f7h8", 1547: "g7a1", 1548: "g7g1", 1549: "g7b2", 1550: "g7g2", 1551: "g7c3",
    1552: "g7g3", 1553: "g7d4", 1554: "g7g4", 1555: "g7e5", 1556: "g7f5", 1557: "g7g5", 1558: "g7h5", 1559: "g7e6",
    1560: "g7f6", 1561: "g7g6", 1562: "g7h6", 1563: "g7a7", 1564: "g7b7", 1565: "g7c7", 1566: "g7d7", 1567: "g7e7",
    1568: "g7f7", 1569: "g7h7", 1570: "g7e8", 1571: "g7f8", 1572: "g7g8", 1573: "g7h8", 1574: "h7b1", 1575: "h7h1",
    1576: "h7c2", 1577: "h7h2", 1578: "h7d3", 1579: "h7h3", 1580: "h7e4", 1581: "h7h4", 1582: "h7f5", 1583: "h7g5",
    1584: "h7h5", 1585: "h7f6", 1586: "h7g6", 1587: "h7h6", 1588: "h7a7", 1589: "h7b7", 1590: "h7c7", 1591: "h7d7",
    1592: "h7e7", 1593: "h7f7", 1594: "h7g7", 1595: "h7f8", 1596: "h7g8", 1597: "h7h8", 1598: "a8a1", 1599: "a8h1",
    1600: "a8a2", 1601: "a8g2", 1602: "a8a3", 1603: "a8f3", 1604: "a8a4", 1605: "a8e4", 1606: "a8a5", 1607: "a8d5",
    1608: "a8a6", 1609: "a8b6", 1610: "a8c6", 1611: "a8a7", 1612: "a8b7", 1613: "a8c7", 1614: "a8b8", 1615: "a8c8",
    1616: "a8d8", 1617: "a8e8", 1618: "a8f8", 1619: "a8g8", 1620: "a8h8", 1621: "b8b1", 1622: "b8b2", 1623: "b8h2",
    1624: "b8b3", 1625: "b8g3", 1626: "b8b4", 1627: "b8f4", 1628: "b8b5", 1629: "b8e5", 1630: "b8a6", 1631: "b8b6",
    1632: "b8c6", 1633: "b8d6", 1634: "b8a7", 1635: "b8b7", 1636: "b8c7", 1637: "b8d7", 1638: "b8a8", 1639: "b8c8",
    1640: "b8d8", 1641: "b8e8", 1642: "b8f8", 1643: "b8g8", 1644: "b8h8", 1645: "c8c1", 1646: "c8c2", 1647: "c8c3",
    1648: "c8h3", 1649: "c8c4", 1650: "c8g4", 1651: "c8c5", 1652: "c8f5", 1653: "c8a6", 1654: "c8b6", 1655: "c8c6",
    1656: "c8d6", 1657: "c8e6", 1658: "c8a7", 1659: "c8b7", 1660: "c8c7", 1661: "c8d7", 1662: "c8e7", 1663: "c8a8",
    1664: "c8b8", 1665: "c8d8", 1666: "c8e8", 1667: "c8f8", 1668: "c8g8", 1669: "c8h8", 1670: "d8d1", 1671: "d8d2",
    1672: "d8d3", 1673: "d8d4", 1674: "d8h4", 1675: "d8a5", 1676: "d8d5", 1677: "d8g5", 1678: "d8b6", 1679: "d8c6",
    1680: "d8d6", 1681: "d8e6", 1682: "d8f6", 1683: "d8b7", 1684: "d8c7", 1685: "d8d7", 1686: "d8e7", 1687: "d8f7",
    1688: "d8a8", 1689: "d8b8", 1690: "d8c8", 1691: "d8e8", 1692: "d8f8", 1693: "d8g8", 1694: "d8h8", 1695: "e8e1",
    1696: "e8e2", 1697: "e8e3", 1698: "e8a4", 1699: "e8e4", 1700: "e8b5", 1701: "e8e5", 1702: "e8h5", 1703: "e8c6",
    1704: "e8d6", 1705: "e8e6", 1706: "e8f6", 1707: "e8g6", 1708: "e8c7", 1709: "e8d7", 1710: "e8e7", 1711: "e8f7",
    1712: "e8g7", 1713: "e8a8", 1714: "e8b8", 1715: "e8c8", 1716: "e8d8", 1717: "e8f8", 1718: "e8g8", 1719: "e8h8",
    1720: "f8f1", 1721: "f8f2", 1722: "f8a3", 1723: "f8f3", 1724: "f8b4", 1725: "f8f4", 1726: "f8c5", 1727: "f8f5",
    1728: "f8d6", 1729: "f8e6", 1730: "f8f6", 1731: "f8g6", 1732: "f8h6", 1733: "f8d7", 1734: "f8e7", 1735: "f8f7",
    1736: "f8g7", 1737: "f8h7", 1738: "f8a8", 1739: "f8b8", 1740: "f8c8", 1741: "f8d8", 1742: "f8e8", 1743: "f8g8",
    1744: "f8h8", 1745: "g8g1", 1746: "g8a2", 1747: "g8g2", 1748: "g8b3", 1749: "g8g3", 1750: "g8c4", 1751: "g8g4",
    1752: "g8d5", 1753: "g8g5", 1754: "g8e6", 1755: "g8f6", 1756: "g8g6", 1757: "g8h6", 1758: "g8e7", 1759: "g8f7",
    1760: "g8g7", 1761: "g8h7", 1762: "g8a8", 1763: "g8b8", 1764: "g8c8", 1765: "g8d8", 1766: "g8e8", 1767: "g8f8",
    1768: "g8h8", 1769: "h8a1", 1770: "h8h1", 1771: "h8b2", 1772: "h8h2", 1773: "h8c3", 1774: "h8h3", 1775: "h8d4",
    1776: "h8h4", 1777: "h8e5", 1778: "h8h5", 1779: "h8f6", 1780: "h8g6", 1781: "h8h6", 1782: "h8f7", 1783: "h8g7",
    1784: "h8h7", 1785: "h8a8", 1786: "h8b8", 1787: "h8c8", 1788: "h8d8", 1789: "h8e8", 1790: "h8f8", 1791: "h8g8",
    1792: "a7a8q", 1793: "a7a8r", 1794: "a7a8b", 1795: "a7b8q", 1796: "a7b8r", 1797: "a7b8b", 1798: "b7a8q", 1799: "b7a8r",
    1800: "b7a8b", 1801: "b7b8q", 1802: "b7b8r", 1803: "b7b8b", 1804: "b7c8q", 1805: "b7c8r", 1806: "b7c8b", 1807: "c7b8q",
    1808: "c7b8r", 1809: "c7b8b", 1810: "c7c8q", 1811: "c7c8r", 1812: "c7c8b", 1813: "c7d8q", 1814: "c7d8r", 1815: "c7d8b",
    1816: "d7c8q", 1817: "d7c8r", 1818: "d7c8b", 1819: "d7d8q", 1820: "d7d8r", 1821: "d7d8b", 1822: "d7e8q", 1823: "d7e8r",
    1824: "d7e8b", 1825: "e7d8q", 1826: "e7d8r", 1827: "e7d8b", 1828: "e7e8q", 1829: "e7e8r", 1830: "e7e8b", 1831: "e7f8q",
    1832: "e7f8r", 1833: "e7f8b", 1834: "f7e8q", 1835: "f7e8r", 1836: "f7e8b", 1837: "f7f8q", 1838: "f7f8r", 1839: "f7f8b",
    1840: "f7g8q", 1841: "f7g8r", 1842: "f7g8b", 1843: "g7f8q", 1844: "g7f8r", 1845: "g7f8b", 1846: "g7g8q", 1847: "g7g8r",
    1848: "g7g8b", 1849: "g7h8q", 1850: "g7h8r", 1851: "g7h8b", 1852: "h7g8q", 1853: "h7g8r", 1854: "h7g8b", 1855: "h7h8q",
    1856: "h7h8r", 1857: "h7h8b"
}

DICT_1858_MOVES_UCI_TO_INDEX:dict[str,int] = {v: k for k, v in DICT_1858_MOVES_INDEX_TO_UCI.items()}

def get_uci_move_by_index(idx:int) -> str:
    """
    All possible moves in UCI notation from white's perspective.
    With black promotions the number of moves becomes 1858 + (6 * 3 + 2 * 2) * 5 = 1968.
    For engines, the board is always from white's perspective to reduce complexity.
    """
    return DICT_1858_MOVES_INDEX_TO_UCI[idx]

def get_uci_move_index(uci:str) -> int:
    return DICT_1858_MOVES_UCI_TO_INDEX[uci]
